/**  File:        CADstuff.h
 **  Package:     none
 **  Created by:  markan
 **  Modified by: markan, wml
 **  Description: C++ datastructure for UG files
 **
 **  More info about classes:
 **
 **  There is a class corresponding to each entity found
 **  in a Unigrafix file, or at least the ones we have come
 **  across in BMG so far.  ug_CADSpace represents a
 **  UG file as a whole, although it is not bound to a
 **  file name.  It maintains lists of all the following
 **  objects related to that geometrical space:
 **
 **      ug_vert's     (vertices)
 **      ug_edge's
 **      ug_wire's     (wires and faces)
 **      ug_sel's
 **      ug_layer's    (colors)
 **      ug_def's      (definitions)
 **      ug_instance's (instances)
 **      ug_inline's   (inlined files)
 **
 **  and is responsible for their destruction when the
 **  ug_CADSpace itself is deleted.  All these objects
 **  need to be created and destroyed through their
 **  respective CADSpaces.
 **
 **  Some of these classes need special explanation.
 **
 **  ug_edge is an edge connecting two vertices.  Edges
 **  are purely for ease of programming - THEY ARE NOT
 **  RENDERED OR SAVED.  Each wire is guaranteed to
 **  maintain a list of edges that comprise it, each
 **  vertex is guaranteed to maintain a list of edges
 **  it is part of, and vice versa.  Each edge can be
 **  part of no more than one wire, indicated by its
 **  variable "parent."  This is set to NULL if the edge
 **  has no parent.
 **
 **  ug_sel is a collection of wires.  Its lists of
 **  vertices and edges are simply the union of all
 **  such objects contained in these wires, and are
 **  automatically maintained.
 **
 **  Every vertex, wire, face, instance, and inline is
 **  considered to be part of some definition.  If an
 **  object is not inside an actual "def X; ... end;"
 **  pair in the UG file (in other words, if it is
 **  an entity rendered without some other instance
 **  using it), it is part of the ug_def "nullRegion"
 **/

#ifndef _CADSTUFF_H_
#define _CADSTUFF_H_

#include <string>
#include <list>
#include <vector>
#include <utility>
#include "support.h"
#include "gpc.h"

using namespace std;


#define MAXNAMELENGTH 80           // Amount of space that will be allocated for names
#define TOL 0.001                  // Vertices this close are assumed to be the same
#define NEARTOL 6
#define ug_region ug_def           // For backwards compatibility

/** enum's - see ug_taggable **/
enum tagNames {BLANK, PAIRED, ORIENTED, COORDINDEX, USED, COVERED, PORTAL};
enum tagValues {};

class ug_taggable;
class ug_vector;
class ug_vert;
class ug_edge;
class ug_wire;
class ug_sel;
class ug_layer;
class ug_CADSpace;
class ug_def;
class ug_instance;
class ug_inline;
class ug_vertOfWire;
class twovector;

/** Class declarations **/

/** ug_taggable
 **
 ** This is a class from which vertices and wires descend.
 ** Its purpose is to allow them to be tagged with name-
 ** value pairs, which are enumerated above.
 **/

class ug_taggable {
 public:

  /*
  list<int> names;
  list<int> values;
  */

  list< pair<int,int> > names_values;

  int check(int name);
  void tag(int name, int value);

 protected:
  class first_equals {
  public:
    first_equals(int n) : name(n) {}
    bool operator ()(pair<int,int> p) { return p.first == name; }
  protected:
    int name;
  };

};

/** ug_vector
 **
 ** Sometimes it's useful just to have a vector around...
 **/

class ug_vector {
 public:
  double x, y, z;

  ug_vector(double a, double b, double c) { x = a; y = b; z = c; }
};


/** ug_vert - represents a vertex
 **/

class ug_vert : public ug_taggable {
 private:

  // this is really unsafe
  //char name[MAXNAMELENGTH];                // Name - must be unique

  string name;                             // Name - must be unique

 public:
  double x, y, z;
  list<ug_edge *> aEdges;                  // Adjacent edges
  list<ug_wire *> pWires;                  // Wires containing it
  ug_CADSpace *home;                       // ug_CADSpace containing it
  ug_def *myRegion;                        // Definition of which it is part

  /** Public methods - see implementation for details **/
  ug_vert(ug_CADSpace *where, double xc, double yc, double zc, ug_def *);
  ug_vert(ug_CADSpace *where, char *n, double xc, double yc, double zc, ug_def *);
  bool match(ug_vert *v);
  char *get_name();
  double distance(ug_vert *other);
  ug_vert *dup(double dx, double dy, double dz, ug_def *);
  ug_vert *dupUsingAxis(ug_edge *axis, double xc, double yc, ug_def *);
  void add(ug_edge *e);
  void add(ug_wire *w);
  void set (double xc, double yc, double zc);
  ~ug_vert();
};

/** ug_edge - represents an edge
 **/

class ug_edge {
 public:
  ug_vert *aVerts[2];                      // Adjacent vertices
  ug_wire *parent;                         // ug_wire containing it, NULL if none
  ug_CADSpace *home;

  /** Public methods - see implementation for details **/
  ug_edge(ug_CADSpace *, ug_vert *v1, ug_vert *v2, ug_wire *parent);
  bool coincides(ug_edge *other);
  bool contains(ug_vert *);
  bool isAbove(ug_vert *v);
  bool isParallel(ug_edge *other);
  bool matchDir(ug_edge *other);
  bool overlaps(ug_edge *other);
  double distance(ug_edge *other);
  double distance(ug_vert *);
  double length();
  double project(ug_vert *v);
  ug_edge *dup();
  ug_edge *overlap(ug_edge *other);
  ug_wire *extrude(double s, double h, ug_def *d);
  void flip();
};

/** ug_wire - represents a wire or a face
 **
 ** No three consecutive vertices may be collinear, no two may be equal.
 ** The class will correct both of these things when vertices are added.
 ** To make the wire a face, use the close() method - do not include the
 ** last vertex twice.  Faces must have at least 3 vertices, all coplanar,
 ** although the coplanar condition is not (yet) checked.  Finally,
 ** self intersecting and non-convex polygons are ok.
 **/

class ug_wire : public ug_taggable {
 public:
  bool isFace;                       // true = face; false = wire
  char name[MAXNAMELENGTH];          // name; must be unique
  h_table<ug_vert> vertList;         // List of vertices in the wire
  //list<ug_edge *> edgeList;          // List of edges in the wire
  vector<ug_edge*> edgeList;
  ug_CADSpace *home;                 // ug_CADSpace containing it
  ug_layer *myLayer;                 // Its color, NULL if none
  ug_def *myRegion;                  // ug_def containing it

  /** Public methods - see implementaiton for details **/
  ug_wire(char *name);
  ug_wire(ug_CADSpace *where, ug_layer *layer, ug_def *);

  bool matchOrient(ug_wire *w);
  char *get_name();
  double area();
  double cmx();
  double cmy();
  double length();
  int numEdges();
  int numVerts();
  ug_edge *getEdge(int i);
  ug_vert *firstJoin(ug_wire *, bool *);
  ug_vert *firstSplit(ug_wire *, bool);
  ug_vert *getVert(int i);
  ug_sel *extrude(double s, double h, ug_def *d);
  ug_wire *dup(double xc, double yc, double zc, ug_def *d); 
  ug_wire *dupClean(ug_def *d);
  ug_wire *unfill();

  void replaceVert(ug_vert *preV, ug_vert *newV);

  void close();
  void flip();
  void open();
  void printme();
  void pushVert(ug_vert *v);
  void setLayer(ug_layer *layer);
  void split(ug_vert *, ug_wire **, ug_wire **);
  void writeGPC(gpc_vertex_list *);
};

// ** Selection (i.e. a collection of wires)

class ug_sel {
 public:
  h_table<ug_vert> vertList;
  //list<ug_edge *> edgeList;
  vector<ug_edge *> edgeList;
  h_table<ug_wire> wireList;
  char name[MAXNAMELENGTH];

  ug_sel();
  ug_sel(char *name);
  ug_sel(gpc_polygon *, ug_CADSpace *, ug_layer *, ug_def *, double z);

  void add(ug_wire *w);
  void merge(ug_sel *m);
  void writeGPC(gpc_polygon *p);

  int numVerts();
  int numEdges();
  int numWires();
  ug_vert *getVert(int i);
  ug_edge *getEdge(int i);
  ug_wire *getWire(int i);
  ug_sel *dup(double, double, double);

  void setLayer(ug_layer *l);
  //  void setRegion(ug_region *r);

  //  ug_CADSpace *toCADSpace();
  ug_sel *clipToLayer(ug_layer *l);
  ug_sel *clipAboveSize(double);
  void writeBBox(double *, double *, double *, double *);
  ug_vertOfWire anyMatchingVert(ug_vert *v, ug_wire *except);
  void fOrientWires(ug_vert *v, ug_wire *b);
  void bOrientWires(ug_vert *v, ug_wire *b);
  void matchOrient(ug_sel *s);
  void align(ug_sel *, double *, double *);
  int evalAlign(ug_sel *, double, double);
  bool covers(ug_vert *);
  ug_vert *nearestVert(ug_vert *v);
  void writeVRML(FILE *);
  void writeIV(FILE *);
};

// ** CADSpace

class ug_CADSpace {
 private:
  int highestUsedVertNum;
  int highestUsedWireNum;

  int highestUsedInstanceNum;

  h_table<ug_vert> vertList;
  //list<ug_edge *> edgeList;
  h_table<ug_wire> wireList;
  list<ug_instance *> instanceList;
  list<ug_inline *> inlineList;
  h_table<ug_layer> layerList;
  //list<ug_region *> regionList;
  vector<ug_region*> regionList;

  ug_region *nullRegion;

 public:
  //ug_def *nullRegion;
  //  ug_wire *uWire;
  
  // this is here for convenience -- it is never even used normally
  vector <ug_edge*>edgeList;

  ug_CADSpace();

  ug_vert *addNewVert(double x, double y, double z, ug_def *);
  ug_vert *addNewVert(char *name, double x, double y, double z, ug_region *region);
  ug_wire *addNewWire(ug_layer *layer, ug_def *);
  ug_wire *addNewWire(gpc_vertex_list *, ug_layer *, ug_def *, double);
  ug_wire *newWire(char *name);
  ug_wire *addNewFace(ug_layer *layer, ug_vert *v1, ug_vert *v2, ug_vert *v3, ug_vert *v4, ug_def *);
  ug_layer *addNewLayer(char *name, double r, double g, double b, int color);

  ug_layer *addExistingLayer(ug_layer *layer);

  ug_region *addNewRegion(char *name);
  void include(char *filename, ug_region *region);
  ug_instance *addNewInstance(ug_def *d);
  void addWire(ug_wire *w, ug_layer *layer, ug_region *region);

  // this is similar to addNewVert() with a name,
  //   but it mangles it slightly for safety --
  //   using addNewVert() with a name, right now, is unsafe
  //   because of possible name collisions
  ug_vert *addNewNamedVert(char *name, double x, double y, double z,
                           ug_region *region);


  // removes all instances, and replaces them with the actual
  //   structures in the referenced region
  void flatten();

  int numWires();
  int numVerts();
  int numLayers();
  int numRegions();
  ug_vert *getVertByName(char *name);
  ug_vert *getVert(int i);
  ug_wire *getWire(int i);
  ug_layer *getLayer(char *name);
  ug_layer *getLayerByNumber(int i);
  ug_region *getRegion(char *name);
  ug_region *getRegionByNumber(int i);

  ug_region *getNullRegion();

  int unusedVertNum();
  int unusedWireNum();
  int unusedInstanceNum();
  void colorFirstColorless(int i);

  ug_sel *selectAll();

  void print_stats();
  void writeVRML(FILE *);
  void writeIV(FILE *);

};

// ** Just a layer class

class ug_layer {
 public:
  char name[MAXNAMELENGTH];
  double r, g, b;
  int c;

  ug_layer(char *n, double rc, double gc, double bc, int color);
  char *get_name() { return name; }
};


class ug_region {
 private:
  h_table<ug_vert> vertList;
  h_table<ug_wire> wireList;
  //list<ug_instance *> instanceList;
  vector<ug_instance*> instanceList;
  //list<ug_inline *> inlineList;
  vector<ug_inline*> inlineList;

 public:
  char name[MAXNAMELENGTH];

  ug_region(char *n);

  int numVerts();
  int numWires();
  int numInstances();
  int numInlines();
  ug_instance *getInstance(int);
  ug_inline *getInline(int);
  ug_vert *getVert(int i);
  ug_wire *getWire(int i);
  void add(ug_vert *v);
  void add(ug_wire *w);
  void add(ug_instance *i);
  void add(ug_inline *i);
  void addSelfToCADSpace(ug_CADSpace *);
};

class ug_vertOfWire {
 public:
  bool found;
  int index;
  ug_vert *v;
  ug_wire *w;

  ug_vertOfWire() { found = false; }
};

/* class portal {
 public:
  ug_wire *jambs[2];
  bool isWindow;
  bool outside[2];
  portal(ug_wire *j1, ug_wire *j2, bool iw, bool o1, bool o2) {
    jambs[0] = j1; jambs[1] = j2; isWindow = iw; outside[0] = o1; outside[1] = o2;
  }
};

class floorModel {
 public:
  double zBase, zFloor, zWindowB, zWindowT, zCeiling, zTop, zDoorT;
  ug_sel *solids, *inWalls, *outWalls, *floors, *shafts, *verbatim, *actualDoors, *ceilings, *skinSel;
  list<portal *> portals;

  floorModel();
  ~floorModel();
  void interpret(ug_sel *plan, ug_CADSpace *where);
  ug_sel *extrude(ug_CADSpace *where);
  ug_sel *extrudeLowLOD(ug_CADSpace *where, double);
  }; */

class ug_instance {
 public:
  ug_region *basedOn;
  ug_region *myRegion;
  ug_layer *myLayer;
  ug_CADSpace *home;
  //char transform[MAXNAMELENGTH];
  //char name[MAXNAMELENGTH];
  string transform;
  string name;

  ug_instance(ug_CADSpace *, ug_def *);
  ug_instance(ug_CADSpace *, ug_region *, ug_region *, double, double, double, double, double, double, double);
  void setRegion(ug_region *);
  void setObject(ug_region *);
  void setLayer(ug_layer *);
  ug_region *getRegion();
  ug_layer *getLayer();
  ug_region *getObject();
  void makeTransform(char *);
  void setName(char *);
  void easyTransform(double, double, double, double, double, double, double);
  const char *getTransform();
  const char *getName();
  char *get_name() { return (char*)name.c_str(); }
};

// for including other files
class ug_inline {
 public:
  char filename[MAXNAMELENGTH];
  ug_region *myRegion;
  ug_region *basedOn;
  ug_CADSpace *home;
  char name[MAXNAMELENGTH];

  ug_inline(ug_CADSpace *, ug_region *, char *);
  void setRegion(ug_region *);
  void setFilename(char *);
  ug_region *getRegion();
  char *getFilename();
  char *get_name() { return name; }
};

/*************************************************************************/
/**                        FUNCTION DECLARATIONS                        **/
/*************************************************************************/

void init_datastructures();

/**this function determines whether two line segments (ug_edges) intersect */

bool intersect(const ug_edge* edge1, const ug_edge* edge2);

/** determines whether two wires overlap horizontally */
bool intersect(ug_wire* wire1, ug_wire* wire2);

/**determines if a ug_vert is contained within a ug_wire (assumed to
   be a closed polygon */

bool inside(ug_vert* vertex, ug_wire* wire);

class twovector {
 public:
  double x, y;
  bool operator== (twovector a) {
    if (a.x == x && a.y == y) return true;
    return false;
  }
};



#endif
