%{

/**  File:        ug_read.y --> ug_read.tab.c
 **  Created by:  wml
 **  Modified by: wml
 **  Invocation:  bison -p ug_read -d ug_read.y
 **  Description: Unigraphics file parser.  Reads a ug file into a data-
 **    structure and returns.
 **/

#include <iostream>
#include <stdio.h>
#include <string.h>
#include <assert.h>
#include <malloc.h>

  //#include <list.h>

#include "../CADlib2/CADstuff.h"
#include "ug_write.h"

/**  struct:      ug_read_str
 **  Description: Allows for a list of strings to be held while reading in data
 **    from the file.  This is mainly designed to keep a list of vertices for a
 **    wire or face as it's being read in.
 **/
struct ug_read_str{
  char *str;            /** Data. **/
  ug_read_str *next;    /** Pointer to the next string in the list. **/
};

/**  struct:      ug_read_double
 **  Description: Linked list of doubles.  Used for keeping track of the points
 **    of a graphic to map onto a face.
 **/
struct ug_read_double{
  double x, y;          /** X, Y coordinates. **/
  ug_read_double *next; /** Next pair in the list. **/
};

/** External variables. **/
extern unsigned int line;    /** Line counter. **/
extern FILE *ug_readin;      /** Input file pointer. **/

ug_CADSpace *ucs;            /** Structure to store file into. **/
char *region_name;           /** Current region. **/
int unnamed_unit;            /** Number of unnamed wire/face. **/

/** Lexical Analyzer **/
extern int ug_readlex();

/**  Function:    ug_readerror(*str)
 **  Description: yyerror(*str), modified for the purposes of this project.
 **/
void ug_readerror(char *str)
{
  cerr << "UG_READ Error: " << str << endl;
  return;
}

%}

%union{
  double num;                 /** All numbers stored as double floats. **/
  char *str;                  /** Strings, identifiers, etc. **/
  struct ug_read_str *urs;    /** NULL-terminated list of strings. **/
  struct ug_read_double *urd; /** NULL-terminated list of doubles. **/
}

%token TOK_LAYER TOK_DEFINITION TOK_END TOK_VERT TOK_WIRE TOK_FACE TOK_INSTANCE
%token TOK_OC TOK_CC TOK_OS TOK_CS TOK_OP TOK_CP TOK_EOL TOK_COLOR TOK_INCLUDE
%token <num> DOUBLE
%token <str> TRANSFORM IDENT ARRAY

%type <str> Iident Icolor Itransform WireColor
%type <urs> VertList
%type <urd> GraphicCoords

%%

/**  Unigraphics files are parsed one line at a time; that is, definitions of
 **  things do not span multiple lines.
 **/
file:		/** Nothing. **/
		| file line
		;

/**  See each type of line for a description.
 **/
line:		LayerData
		| RegionDef
		| RegionEnd
		| VertexDef
		| WireDef
		| InstanceDef
		| ArrayDef
		| ColorDef
		| IncludeStmt
		| EscapeText
		| CommentStmt
		| error '\n' {
		  fprintf(stderr, "Recovering from error on line: %d\n", line);
		  yyerrok;
		}
		;

/**  Comments in Unigraphix use the { } symbols.
 **/
CommentStmt:	TOK_OC CommentInter TOK_CC { fprintf(stderr, "Got a comment\n"); }
		;

/**  Used with CommentStmt.
 **/
CommentInter:	/** Nothing. **/
		| CommentInter CommentStmt { }
		| CommentInter IDENT { free($2); }
		| CommentInter DOUBLE { }
		| CommentInter TRANSFORM { free($2); }
		;

/**  I don't know what this is, or how it differs from a comment, but I collect
 **  it, just in case.
 **/
EscapeText:	TOK_OP EscapeInternal TOK_CC { fprintf(stderr, "Got an escape\n"); }
		;

/**  Helper to EscapeText.
 **/
EscapeInternal: /** Nothing. **/
		| EscapeInternal EscapeText { }
		| EscapeInternal DOUBLE { }
		| EscapeInternal IDENT { free($2); }
		| EscapeInternal TRANSFORM { free($2); }
		;

/**  Create a new layer.  This is to sort out wires and faces, and is often
 **  represented by a variance of color.  TOK_LAYER == 'c_rgb'.
 **/
LayerData:	TOK_LAYER IDENT DOUBLE DOUBLE DOUBLE TOK_EOL {
                  ucs->addNewLayer($2, $3, $4, $5, -1);
		  delete[] $2;
		}
		;

/**  Not quite sure what an array is in this context.
 **/
ArrayDef:	ARRAY {
		  cerr << "Array: " << $1 << endl;
		  delete[] $1;
		}
		;

/**  The color number of a layer is not necessarily at the end of a line, but
 **  may actually be at the end of the file (poor choice by UG, in my opinion).
 **  This means that a layer may not have an official color until the end of
 **  the file, though the colors must appear in the same order as the layer
 **  declarations.
 **/
ColorDef:	TOK_OC TOK_COLOR DOUBLE TOK_CC {
		  ucs->colorFirstColorless( (int) $3);
		}
		;

/**  Include the file, represented by $2 in the model.
 **/
IncludeStmt:	TOK_INCLUDE IDENT TOK_EOL {
		  ucs->include($2, ucs->getRegion(region_name));
		  free($2);
		}
		;

/**  Define a new region and set everything defined within this region to it.
 **/
RegionDef:	TOK_DEFINITION IDENT TOK_EOL {
		  assert(region_name == NULL);
		  region_name = $2;
		  ucs->addNewRegion(region_name);
		}
		;

/**  At the end of each region definition, there is a string that looks like so
 **  "end;".  When this string is reached, end the current region and set the
 **  region_name to NULL.
 **/
RegionEnd:	TOK_END TOK_EOL {
		  assert(region_name != NULL);
		  delete[] region_name;
		  region_name = NULL;
		}
		;

/**  Define a vertex and put it in the global CADSpace file.
 **/
VertexDef:	TOK_VERT IDENT DOUBLE DOUBLE DOUBLE TOK_EOL {
		  /** 3D case. **/
		  ucs->addNewVert($2, $3, $4, $5, ucs->getRegion(region_name));
		  delete[] $2;
		}
		| TOK_VERT IDENT DOUBLE DOUBLE TOK_EOL {
		  /** 2D case. **/
		  ucs->addNewVert($2, $3, $4, 0, ucs->getRegion(region_name));
		  delete[] $2;
		}
		;

/**  Define a wire.  In UG, wires have a variable number of vertices, so I have
 **  to make a list of the vertices for when I create the wire (as they are
 **  each evaluated before this segment of code).  Also, the 'wire' in question
 **  may also be a face.
 **/
WireDef:	TOK_WIRE WireColor TOK_OP VertList TOK_CP WireColor TOK_EOL {
		  char issue_buff[256];
		  ug_read_str *urs1, *urs2 = NULL;
		  ug_wire *w;

		  if($4 == NULL){
		    /** If there are no vertices, report the error. **/
		    cerr << "Empty wire: " << $2 << endl;
		  }else{

		    /** Some wires and faces are unnamed.  Give them a default
		        name. **/
		    if($2 == NULL){
		      sprintf(issue_buff, "unnamed_unit_%d", unnamed_unit);
		      ++unnamed_unit;
		      $2 = strdup(issue_buff);
		    }

		    /** Get the next wire from the datastructure. **/
		    w = ucs->newWire($2);
		    for(urs1 = $4; urs1 != NULL; urs1 = urs1->next){
		      /** For each vertex, add it to the list in the wire. **/
		      if(urs2 != NULL){
			delete[] urs2->str;
			delete urs2;
		      }
		      urs2 = urs1;
		      w->pushVert(ucs->getVertByName(urs1->str));
		    }

		    /** Add the wire to the datastructure. **/
		    if(ucs->getLayer($6) == NULL){
		      /** There was no such layer. **/
		      if($6 == NULL){
		        sprintf(issue_buff, "Wire %s had no layer.  %s",
			        $2, "Adding it to EMPTY_LAYER");
			ug_readerror(issue_buff);
		      }else{
		        sprintf(issue_buff, "Wire %s had a bad layer.  %s",
			        $2, "Adding it to EMPTY_LAYER");
			ug_readerror(issue_buff);
			free($6);
		      }
		      $6 = strdup("EMPTY_LAYER");
		      if(ucs->getLayer($6) == NULL)
		        ucs->addNewLayer($6, 0.500, 0.500, 0.500, -1);
		    }
		    ucs->addWire(w, ucs->getLayer($6),
				 ucs->getRegion(region_name));
		  }
		  delete[] $6;
		  delete[] $2;

		}
		| TOK_FACE WireColor TOK_OP VertList TOK_CP WireColor TOK_EOL {
		  char issue_buff[256];
		  ug_read_str *urs1, *urs2 = NULL;
		  ug_wire *f;

		  if($4 == NULL){
		    /** If there are no vertices, report the error. **/
		    cerr << "Empty face: " << $2 << endl;
		  }else{

		    /** Some wires and faces are unnamed.  Give them a default
		        name. **/
		    if($2 == NULL){
		      sprintf(issue_buff, "unnamed_unit_%d", unnamed_unit);
		      ++unnamed_unit;
		      $2 = strdup(issue_buff);
		    }

		    /** Get the next face from the datastructure. **/
		    f = ucs->newWire($2);
		    for(urs1 = $4; urs1 != NULL; urs1 = urs1->next){
		      /** For each vertex, add it to the list in the wire. **/
		      if(urs2 != NULL){
			delete[] urs2->str;
			delete urs2;
		      }
		      urs2 = urs1;
		      f->pushVert(ucs->getVertByName(urs1->str));
		    }

		    /** Turn the wire into a face (closed polygon). **/
		    f->close();
		    /** Add the wire to the datastructure. **/
		    if(ucs->getLayer($6) == NULL){
		      /** There was no such layer. **/
		      if($6 == NULL){
		        sprintf(issue_buff, "Face %s had no layer.  %s",
			        $2, "Adding it to EMPTY_LAYER");
			ug_readerror(issue_buff);
		      }else{
		        sprintf(issue_buff, "Face %s had a bad layer.  %s",
			        $2, "Adding it to EMPTY_LAYER");
			ug_readerror(issue_buff);
			free($6);
		      }
		      $6 = strdup("EMPTY_LAYER");
		      if(ucs->getLayer($6) == NULL)
		        ucs->addNewLayer($6, 0.500, 0.500, 0.500, -1);
		    }
		    ucs->addWire(f, ucs->getLayer($6),
				 ucs->getRegion(region_name));
		  }
		  delete[] $6;
		  delete[] $2;

		}
		| TOK_FACE WireColor TOK_OP VertList TOK_CP WireColor TOK_OS
		  GraphicCoords TOK_CS TOK_EOL {

		  char issue_buff[256];
		  ug_read_str *urs1, *urs2 = NULL;
		  ug_wire *f;

		  if($4 == NULL){
		    /** If there are no vertices, report the error. **/
		    cerr << "Empty face: " << $2 << endl;
		  }else{

		    /** Some wires and faces are unnamed.  Give them a default
		        name. **/
		    if($2 == NULL){
		      sprintf(issue_buff, "unnamed_unit_%d", unnamed_unit);
		      ++unnamed_unit;
		      $2 = strdup(issue_buff);
		    }

		    /** Get the next face from the datastructure. **/
		    f = ucs->newWire($2);
		    for(urs1 = $4; urs1 != NULL; urs1 = urs1->next){
		      /** For each vertex, add it to the list in the wire. **/
		      if(urs2 != NULL){
			delete[] urs2->str;
			delete urs2;
		      }
		      urs2 = urs1;
		      f->pushVert(ucs->getVertByName(urs1->str));
		    }

		    /** Turn the wire into a face (closed polygon). **/
		    f->close();
		    /** Add the wire to the datastructure. **/
		    if(ucs->getLayer($6) == NULL){
		      /** There was no such layer. **/
		      if($6 == NULL){
		        sprintf(issue_buff, "Face %s had no layer.  %s",
			        $2, "Adding it to EMPTY_LAYER");
			ug_readerror(issue_buff);
		      }else{
		        sprintf(issue_buff, "Face %s had a bad layer.  %s",
			        $2, "Adding it to EMPTY_LAYER");
			ug_readerror(issue_buff);
			free($6);
		      }
		      $6 = strdup("EMPTY_LAYER");
		      if(ucs->getLayer($6) == NULL)
		        ucs->addNewLayer($6, 0.500, 0.500, 0.500, -1);
		    }
		    ucs->addWire(f, ucs->getLayer($6),
				 ucs->getRegion(region_name));
		  }
		  delete[] $6;
		  delete[] $2;

		}
		;

/**  Make a linked list of ug_read_str's in the same order as the vertices in
 **  a wire's list are found.
 **/
VertList:	{ $$ = NULL; }
		| IDENT VertList {
		  ug_read_str *s;
		  s       = new ug_read_str;
		  s->str  = $1;
		  s->next = $2;
		  $$      = s;
		}
		;

/**  This is instnace information.  It is an instantiation of a region, here
 **  specified by $4.  It may or may not have other information such as a
 **  transform and such.
 **/
InstanceDef:	TOK_INSTANCE Iident TOK_OP IDENT Icolor Itransform TOK_CP
		TOK_EOL {

		  ug_instance *i;
		  i = ucs->addNewInstance(ucs->getRegion(region_name));
		  if($2 != NULL){
		    i->setName($2);
		    free($2);
		  }

		  i->setObject(ucs->getRegion($4));
		  free($4);

		  if($5 != NULL){
		    i->setLayer(ucs->getLayer($5));
		    free($5);
		  }

		  if($6 != NULL){
		    i->makeTransform($6);
		    free($6);
		  }

		}
		;

/**  Each coordinate corresponds to a vertex in the list of vertices in a face.
 **  This point is a location in the graphic that gets mapped to it.
 **/
GraphicCoords:	{
		  $$ = NULL;
		}
		| DOUBLE DOUBLE GraphicCoords {
		  $$ = new ug_read_double;
		  $$->x    = $1;
		  $$->y    = $2;
		  $$->next = $3;
		}
		;

/**  Three optional parts of an instance.  See the Instance description for
 **  more details.
 **/
Iident:		/** Nothing. **/ { $$ = NULL; }
		| IDENT { $$ = $1; }
		;
Icolor:		/** Nothing. **/ { $$ = NULL; }
		| IDENT { $$ = $1; }
		;
Itransform:	/** Nothing. **/ { $$ = NULL; }
		| TRANSFORM { $$ = $1; }
		;

WireColor:	/** Nothing. **/ { $$ = NULL; }
		| IDENT { $$ = $1; }
		;

%%

/**  Function:    ug_parse(&uc, *filename)
 **  Description: Use this function, rather than calling ug_readparse()
 **    directly.  This sets up all of the defaults and prepares the
 **    datastructure for reading.
 **/
void ug_parse(ug_CADSpace &uc, char *filename)
{
  //int i;
  //ug_region *ur;

  region_name = NULL;
  line = 0;
  ucs = &uc;
  unnamed_unit = 0;

  ug_readin = fopen(filename, "r");
  if(ug_readin != NULL){
    cerr << "Reading from file: " << filename << endl;
    ug_readparse();
    cerr << "Done reading file." << endl;
    fclose(ug_readin);
  }else{
    cerr << "Couldn't read from file: " << filename << endl;
  }

  if(unnamed_unit > 0){
    cerr << "Named " << unnamed_unit << " wires and faces." << endl;
  }

  return;
}

