/**  File:        ug_write.h
 **  Package:     none
 **  Created by:  wml
 **  Modified by: wml
 **  Description: Functions designed to write out a building in ug format.  The
 **    various functions write out different aspects of the building, however,
 **    depending on which ones you call.
 **/

#ifndef _UG_WRITE_H_
#define _UG_WRITE_H_

/*****************************************************************************/
/**                              HEADER FILES                               **/
/*****************************************************************************/

#include <stdio.h>
#include <string.h>
#include <assert.h>
#include "../CADlib2/CADstuff.h"


// hack
class region_equals {
 public:
  region_equals(ug_region *r) : reg(r) {}
  bool operator ()(ug_region *r) { return reg == r; }

 protected:
  ug_region *reg;
};


/*****************************************************************************/
/**                          FUNCTION DEFINITIONS                           **/
/*****************************************************************************/

/**  Function:    ug_write(&uc, *filename)
 **  Description: Write the ug_CADSpace file to *filename in unigraphics.
 **/
void ug_write(ug_CADSpace &uc, char *filename)
{
  FILE *fp;                   /** The output file pointer. **/
  unsigned int num;
  int num2, num3;        /** Upper bound counters. **/
  unsigned int i;
  int j, k;                /** Iterators. **/
  //list<ug_wire *> *lw, *lf;   /** Lists of wires and faces. **/
  vector<ug_wire*> *lw, *lf;
  ug_layer *ul;               /** Layer/Color pointer. **/
  ug_region *ur;              /** Region pointer. **/
  ug_vert *v;                 /** Vertex pointer. **/
  ug_wire *w;                 /** Wire pointer. **/
  ug_inline *inl;             /** Inline iterator. **/
  ug_instance *inst;          /** Instance iterator. **/
  //list<ug_region*> *rtable;   /** List for ordering regions. **/
  vector<ug_region*> *rtable;

  vector<ug_region*>::iterator it;

  /** Open the file. **/
  fp = fopen(filename, "w");

  if(fp){

    /** Output layer data, first. **/
    num = uc.numLayers();
    for(i = 0; i < num; ++i){
      ul = uc.getLayerByNumber(i);
      assert(ul != NULL);

      fprintf(fp, "c_rgb %s %f %f %f ; { color %d }\n",
              ul->name, (float) ul->r, (float) ul->g, (float) ul->b, ul->c);
    }

    /** Set the regions in order to be output.  This means that if a certain
        region has dependencies on another, it will not be output first. **/
    num = uc.numRegions();
    //rtable = new list<ug_region*>();
    rtable = new vector<ug_region*>();
    k = 1;                         /** Used to prevent an infinite loop. **/
    while (rtable->size() < num && k != 0) {
      //while(rtable->size < num && k != 0){
      k = 0;                       /** Reset the counter. **/
      for(i = 0; i < num; ++i){
        // Go through each region in the file.
        ur = uc.getRegionByNumber(i);
        assert(ur != NULL);

        
        it = find_if(rtable->begin(), rtable->end(), region_equals(ur));

        if (it == rtable->end()) {
          //if(!rtable->contains(ur)){
          num2 = ur->numInstances();
          if(num2 == 0){
            // If there are no dependencies, simply add it.
            //rtable->add(ur);
            rtable->push_back(ur);
            ++k;                    /** Increment insertion counter. **/
          }else{
            num3 = 1;               /** Make sure dependencies have all been
                                        met for this region. **/
            for(j = 0; j < num2 && num3; ++j){
              // Go through each instance to see if its region is in the list.
              assert(ur->getInstance(j)->getObject() != NULL);

              it = find_if(rtable->begin(), rtable->end(),
                           region_equals(ur->getInstance(j)->getObject()));

              if (it == rtable->end()) num3 = 0;
              //if(!rtable->contains(ur->getInstance(j)->getObject())) num3 = 0;
            }
            if(num3){
              rtable->push_back(ur);
              //rtable->add(ur);
              ++k;                  /** Increment insertion counter. **/
            }
          }
        }
      }
    }

    if(k == 0){
      fprintf(stderr, "UG_WRITE ERROR: File contains circular dependencies\n");
      fprintf(stderr, "                Not writing file.");
      return;
    }

    /** Output each region. **/
    //num = rtable->size;
    num = rtable->size();
    for(i = 0; i < num; ++i){
      ur = rtable->at(i);
      assert(ur != NULL);

      if(strcmp(ur->name, "")) fprintf(fp, "def %s;\n", ur->name);

      /** Print out all of the vertices in the region. **/
      num2 = ur->numVerts();
      for(j = 0; j < num2; ++j){
        v = ur->getVert(j);
        assert(v != NULL);

        fprintf(fp, "v %s %f %f %f ;\n", v->get_name(), v->x, v->y, v->z);
      }

      /** Make a list of wires and a list of faces for this region. **/
      num2 = ur->numWires();
      //lw = new list<ug_wire *>;
      lw = new vector<ug_wire*>;
      //lf = new list<ug_wire *>;
      lf = new vector<ug_wire*>;
      for(j = 0; j < num2; ++j){
        w = ur->getWire(j);
        assert(w != NULL);
        if(w->isFace){
          //lf->add(w);
          lf->push_back(w);
        }else{
          //lw->add(w);
          lw->push_back(w);
        }
      }

      /** Add all of the wires to the file. **/
      //num2 = lw->size;
      num2 = lw->size();
      for(j = 0; j < num2; ++j){
        w = lw->at(j);
        assert(w != NULL);

        fprintf(fp, "w %s (", w->name);
        num3 = w->numVerts();
        for(k = 0; k < num3; ++k){
          v = w->getVert(k);
          assert(v != NULL);

          fprintf(fp, " %s ", v->get_name());
        }

        fprintf(fp, ") %s ;\n", (w->myLayer)->name);
      }

      /** Add all of the faces to the file. **/
      //num2 = lf->size;
      num2 = lf->size();
      for(j = 0; j < num2; ++j){
        w = lf->at(j);
        assert(w != NULL);

        fprintf(fp, "f %s (", w->name);
        num3 = w->numVerts();
        for(k = 0; k < num3; ++k){
          v = w->getVert(k);
          assert(v != NULL);

          fprintf(fp, " %s ", v->get_name());
        }

        fprintf(fp, ") %s ;\n", (w->myLayer)->name);
      }

      /** Add any includes and instances to the region. **/

      num2 = ur->numInlines();
      for(j = 0; j < num2; ++j){
        inl = ur->getInline(j);
        fprintf(fp, "include %s;\n", inl->getFilename());
      }

      num2 = ur->numInstances();
      for(j = 0; j < num2; ++j){
        inst = ur->getInstance(j);
        fprintf(fp, "i ");
        if(strcmp(inst->name.c_str(), ""))
          fprintf(fp, "%s", inst->name.c_str());
        fprintf(fp, "(%s", inst->getObject()->name);
        if(inst->getLayer() != NULL)
          fprintf(fp, " %s", inst->getLayer()->name);
        if(strcmp(inst->getTransform(), ""))
          fprintf(fp, " %s", inst->getTransform());
        fprintf(fp, ");\n");
      }

      /** End the region. **/

      if(strcmp(ur->name, "")) fprintf(fp, "end;\n");
    }

    fclose(fp);
    delete rtable;

  }else{
    /** Couldn't open the file. **/
    cerr << "Could not open file " << filename << " for writing." << endl;
  }

  return;
}

/**  Function:    ug_write_far(&uc, *filename)
 **  Description: Write a UG file of a building as it would look from far away
 **    (no doors or windows or internal stuff).
 **/
void ug_write_far(ug_CADSpace &uc, char *filename)
{
  FILE *fp;               /** File pointer for *filename. **/

  fp = fopen(filename, "w");

  if(fp){

  }else{
    cerr << "Could not open file " << filename << " for writing." << endl;
  }

  return;
}

#endif

