#include <assert.h>
#include <CADstuff.h>
#include "ug_builder.h"
#include "dxf_reader.h"

//using namespace _STL;


DXF_Reader::DXF_Reader(UG_Builder *ug_builder)
  : ugb(ug_builder),
    polyline_vert_count(0),
    reading_polyline(false),
    ignoring_polyline(false)
{
  
}

DXF_Reader::~DXF_Reader()
{

}


void DXF_Reader::addLayer(const DL_LayerData& data)
{
  assert(!reading_polyline);

  ugb->addLayer(data.name);
}

void DXF_Reader::addBlock(const DL_BlockData& data)
{
  ug_region *r;
  ug_layer *l;

  assert(!reading_polyline);


  r = ugb->beginDefinition(data.name);
  l = ugb->getLayer(this->getCurrentLayer());

  
  fprintf(stdout, "trying to add a block with region: \"%s\"\n",
          this->getCurrentLayer().c_str());
  

  if (l == NULL) {
    fprintf(stdout, "tried to add a block with an invalid region!\n");
    return;
  }
  // now, invalid blocks will not be in the regions map

  fprintf(stdout, "adding a block with a VALID region\n");

  // might want to check for duplicate region names
  regions[data.name] = pair<ug_region*, ug_layer*>(r, l);
}

void DXF_Reader::endBlock()
{
  assert(!reading_polyline);

  ugb->endDefinition();
}

void DXF_Reader::addPoint(const DL_PointData& data)
{
  assert(!reading_polyline);


  ugb->addAnonymousVertex(data.x, data.y, data.z);
}

void DXF_Reader::addLine(const DL_LineData& data)
{
  ug_vert *v1, *v2;
  ug_wire *w;
  string layer;

  assert(!reading_polyline);


  w = ugb->addAnonymousWire(this->getCurrentLayer());
  if (w == NULL) return;

  // ugb checks that v1 and v2 are valid (non-NULL)
  v1 = ugb->addAnonymousVertex(data.x1, data.y1, data.z1);
  v2 = ugb->addAnonymousVertex(data.x2, data.y2, data.z2);
  w->pushVert(v1);
  w->pushVert(v2);
}

void DXF_Reader::addArc(const DL_ArcData& data)
{
  ug_layer *layer;

  assert(!reading_polyline);

  layer = ugb->getLayer(this->getCurrentLayer());

  if (layer == NULL) {
    fprintf(stdout, "tried to add a circle with an invalid region!\n");
    return;
  }

  ugb->addAnonymousArc(data.cx, data.cy, data.cz,
                       data.radius, data.angle1, data.angle2,
                       layer);
}

void DXF_Reader::addCircle(const DL_CircleData& data)
{
  ug_layer *layer;

  assert(!reading_polyline);

  layer = ugb->getLayer(this->getCurrentLayer());
  
  if (layer == NULL) {
    fprintf(stdout, "tried to add a circle with an invalid region!\n");
    return;
  }

  ugb->addAnonymousCircle(data.cx, data.cy, data.cz, data.radius,
                          layer);
}

void DXF_Reader::addEllipse(const DL_EllipseData& data)
{
  ug_layer *layer;

  assert(!reading_polyline);

  layer = ugb->getLayer(this->getCurrentLayer());

  if (layer == NULL) {
    fprintf(stdout, "tried to add an ellipse with an invalid region!\n");
    return;
  }

  ugb->addAnonymousEllipse(data.cx, data.cy, data.cz,
                           data.mx, data.my, data.mz,
                           data.ratio, data.angle1, data.angle2,
                           layer);
}

void DXF_Reader::addPolyline(const DL_PolylineData& data)
{
  //assert(polyline_vert_count==0);
  assert(!reading_polyline);

  reading_polyline = true;

  // UG_Builder::beginAnonymousWire() checks layer validity
  ignoring_polyline = !ugb->beginAnonymousWire(this->getCurrentLayer());

  fprintf(stdout, "starting POLYLINE (layer \"%s\")\n",
          this->getCurrentLayer().c_str());
  
  




  //polyline_vert_count = data.number;

  //printf("this POLYLINE has %d verts\n", data.number);

  // this should probably never happen
  /*
  if (polyline_vert_count == 0) {
    ugb->endAnonymousWire();
  }
  */
}


void DXF_Reader::addVertex(const DL_VertexData& data)
{
  
  // i'm pretty sure only POLYLINEs have VERTEXs
  assert(reading_polyline);

  //assert(polyline_vert_count > 0);

  if (ignoring_polyline) return;

  ugb->addAnonymousVertexToCurrentWire(data.x, data.y, data.z);
  //polyline_vert_count--;
  

  // might want to check that the layer matches the POLYLINE's layer


  /*
  if (polyline_vert_count == 0) {
    ugb->endAnonymousWire();
  }
  */
}

void DXF_Reader::addInsert(const DL_InsertData& data)
{
  assert(!reading_polyline);


  ugb->addInstance(data.name, data.ipx, data.ipy, data.ipz,
                   data.sx, data.sy, data.sz, data.angle);
}


void DXF_Reader::addMText(const DL_MTextData& data)
{
  string layer_name;

  assert(!reading_polyline);

  printf("the width of \"%s\" is %lf (style \"%s\")\n",
         data.text.c_str(), data.width, data.style.c_str());

  layer_name = getAttributes().getLayer();
  ugb->addTextVertex(data.text, layer_name, data.ipx, data.ipy, data.ipz);
}

void DXF_Reader::addMTextChunk(const char* text)
{
  // this is ignored
}

void DXF_Reader::addText(const DL_TextData& data)
{
  string layer_name;

  assert(!reading_polyline);

  printf("the scale factor of \"%s\" is %lf (style \"%s\")\n",
         data.text.c_str(), data.xScaleFactor, data.style.c_str());
  
  layer_name = getAttributes().getLayer();
  ugb->addTextVertex(data.text, layer_name, data.ipx, data.ipy, data.ipz);
}

void DXF_Reader::addDimAlign(const DL_DimensionData& data,
                             const DL_DimAlignedData& edata)
{
  
}

void DXF_Reader::addDimLinear(const DL_DimensionData& data,
                              const DL_DimLinearData& edata){ }
void DXF_Reader::addDimRadial(const DL_DimensionData& data,
                              const DL_DimRadialData& edata){ }
void DXF_Reader::addDimDiametric(const DL_DimensionData& data,
                                 const DL_DimDiametricData& edata){ }
void DXF_Reader::addDimAngular(const DL_DimensionData& data,
                               const DL_DimAngularData& edata){ }
void DXF_Reader::addDimAngular3P(const DL_DimensionData& data,
                                 const DL_DimAngular3PData& edata){ }
void DXF_Reader::addLeader(const DL_LeaderData& data){ }
void DXF_Reader::addLeaderVertex(const DL_LeaderVertexData& data){ }
void DXF_Reader::addHatch(const DL_HatchData& data){ }
void DXF_Reader::addHatchLoop(const DL_HatchLoopData& data){ }
void DXF_Reader::addHatchEdge(const DL_HatchEdgeData& data){ }

// not sure when this happens
void DXF_Reader::endEntity()
{

}

void DXF_Reader::endSequence()
{
  
  // for now, don't worry about this... SEQEND's may occur for ATTRIB's too,
  //   and it is not worth dealing with those for now
  //assert(reading_polyline);

  if (reading_polyline) {
    if (!ignoring_polyline)
      ugb->endAnonymousWire();
    reading_polyline = false;
  }

  //reading_polyline = false;

  fprintf(stdout, "got end of sequence\n");

  //ugb->endAnonymousWire();
  
}
  
void DXF_Reader::setVariableVector(const char* key,
                                   double v1, double v2, double v3, int code){ }
void DXF_Reader::setVariableString(const char* key,
                                   const char* value, int code){ }
void DXF_Reader::setVariableInt(const char* key, int value, int code){ }
void DXF_Reader::setVariableDouble(const char* key, double value, int code){ }
  

/*
void DXF_Reader::printAttributes()
{

}
*/

const string DXF_Reader::getCurrentLayer() const
{
  return attributes.getLayer();
}
