#include <stdlib.h>
#include <CADstuff.h>
#include <StringPack.h>
#include <math.h>
#include <assert.h>
#include <string>
#include "ug_builder.h"

#define LAYER_LINE_LEN (256)
#define WHITESPACE(c) \
      ( ( (c)==' ' ) || ( (c)=='\r' ) || ( (c)=='\n' ) || ( (c)=='\t' ) )
#define NUMBER(c) ( ( (c) >= '0' ) && ( (c) <= '9' ) )

extern void ug_write(ug_CADSpace &uc, char *filename);


UG_Builder::UG_Builder()
  : ug(),
    current_region(NULL),
    current_color_index(0),
    current_poly_wire(NULL),
    arc_layer(NULL)
    //name_layer(NULL)
{

}

UG_Builder::~UG_Builder()
{

}

ug_region *UG_Builder::beginDefinition(const string &name)
{
  // sanity check, replace with an exception
  assert(current_region == NULL);

  current_region = ug.addNewRegion((char*)name.c_str());
  
  // sanity check, again use something else, exceptions etc.
  assert(current_region != NULL);

  return current_region;
}

/*
  beginDefinition() should have been called prior to this,
  and thus current_region is not NULL
*/
void UG_Builder::endDefinition()
{
  
  // sanity check
  // use something better (exceptions)
  assert(current_region != NULL);

  current_region = NULL;
}

ug_vert *UG_Builder::addAnonymousVertex(double x, double y, double z)
{
  ug_vert *v;

  // if current_region is NULL< it's ok -- CADlib2 handles that
  //   and uses the default toplevel region
  v = ug.addNewVert(x, y, z, current_region);

  // again, use exceptions
  assert(v != NULL);

  return v;
}

/*
  this will return NULL if the layer is not valid
*/
ug_wire *UG_Builder::addAnonymousWire(const string &layer_name)
{
  ug_wire *w;
  ug_layer *layer;

  // check that the layer is valid
  layer = ug.getLayer((char*)layer_name.c_str());
  if (layer == NULL) {
    fprintf(stdout,
            "error (non-fatal): DXF file has a wire on an unsupported layer: \"%s\"\n",
            layer_name.c_str());
    return NULL;
  }

  w = ug.addNewWire(layer, current_region);

  // again use, exceptions
  assert(w != NULL);

  return w;
}

/*
  this gives the wire no layer (NULL)
  
  - returns true if the wire is good (on a valid layer),
    false otherwise -- if false, addAnonymousVertexToCurrentWire() should not
    be called for this wire
*/
bool UG_Builder::beginAnonymousWire(const string layer_name)
{
  ug_layer *layer;

  assert(current_poly_wire==NULL);

  layer = ug.getLayer((char*)layer_name.c_str());
  if (layer == NULL) {
    fprintf(stdout,
            "error (non-fatal): DXF file has a wire on an unsupported layer: \"%s\"\n",
            layer_name.c_str());
    return false;
  }
  
  current_poly_wire = ug.addNewWire(layer, current_region);

  assert(current_poly_wire != NULL);

  return true;
}

void UG_Builder::addAnonymousVertexToCurrentWire(double x, double y, double z)
{
  ug_vert *v;

  assert(current_poly_wire != NULL);
  
  v = ug.addNewVert(x, y, z, current_region);
  assert(v != NULL);

  current_poly_wire->pushVert(v);
}

void UG_Builder::endAnonymousWire()
{
  assert(current_poly_wire != NULL);

  current_poly_wire = NULL;
}

void UG_Builder::addLayer(const string &layer_name)
{
  ug_layer *layer;

  printf("requested to add %s\n", layer_name.c_str());

  layer = valid_layers.get_named((char*)layer_name.c_str());
  
  if (layer == NULL) {
    fprintf(stdout, "error (non-fatal): not going to add an invalid layer\n");
    return;
  }

  ug.addExistingLayer(layer);

  /*
  layer = ug.addNewLayer((char*)layer_name.c_str(), 0.0, 1.0, 0.0,
                         current_color_index++);

  // make this an exception
  assert(layer != NULL);
  */

  printf("just added layer with name \"%s\"\n", layer_name.c_str());
  printf("the cadspace has %d layers\n", ug.numLayers());
}

/*
  region_name is the name of teh region for this instance;
  (ox, oy, oz) is the global point where this instance should be inserted,
  sx, sy, and sz are the scale factors,
  and rz is the amount to rotate about the z-axis
*/
void UG_Builder::addInstance(const string &region_name,
                             double ox, double oy, double oz,
                             double sx, double sy, double sz,
                             double rz)
{
  ug_region *region;
  ug_instance *instance;

  region = ug.getRegion((char*)region_name.c_str());
  
  // this is a non-fatal error
  if (region == NULL) {
    fprintf(stdout, "requested instance of unknown region, \"%s\"\n",
            region_name.c_str());
    return;
  }

  instance = ug.addNewInstance(current_region);
  assert(instance != NULL);

  instance->setObject(region);
  instance->easyTransform(sx, sy, sz, rz, ox, oy, oz);

  printf("this instance has name \"%s\"\n", instance->name.c_str());
}

/*
  - the old approach was to make a very specific structure to represent
    the arc -- now we will use a single vertex, with StringPack
*/

void UG_Builder::addAnonymousArc(double cx, double cy, double cz,
                                 double r, double a1, double a2,
                                 ug_layer *layer)
{
  /*
  ug_wire *w;
  ug_vert *v1, *v2, *v3;

  if (arc_layer == NULL) {
    fprintf(stdout, "error (non-fatal): needed to add an arc, "
                    "but no arc layer available\n");
    return;
  }
  
  w = ug.addNewWire(arc_layer, current_region);  
  assert(w != NULL);

  v2 = ug.addNewVert(cx, cy, cz, current_region);
  v1 = ug.addNewVert(cx + r*cos(a1*M_PI/180.0),
                     cy + r*sin(a1*M_PI/180.0),
                     cz, current_region);
  v3 = ug.addNewVert(cx + r*cos(a2*M_PI/180.0),
                     cy + r*sin(a2*M_PI/180.0),
                     cz, current_region);

  assert(v1 != NULL && v2 != NULL && v3 != NULL);

  w->pushVert(v1);
  w->pushVert(v2);
  w->pushVert(v3);
  */

  ug_vert *v;
  StringPack sp;
  string type;

  /*
  if (arc_layer == NULL) {
    fprintf(stdout, "error (non-fatal): needed to add an arc, "
                    "but no arc layer available\n");
    return;
  }
  */

  type = string("arc");
  //layer = string(arc_layer->get_name());
  sp.addString(string("type"), type);
  sp.addString(string("layer"), layer->get_name());
  sp.addDouble(string("radius"), r);
  sp.addDouble(string("angle1"), a1);
  sp.addDouble(string("angle2"), a2);

  v = ug.addNewNamedVert((char*)sp.serialize().c_str(),
                         cx, cy, cz, current_region);
}

void UG_Builder::addAnonymousCircle(double x, double y, double z, double r,
                                    ug_layer *layer)
{
  /*
  ug_wire *w;
  ug_vert *v1, *v2, *v3, *v4;
  
  if (arc_layer == NULL) {
    fprintf(stdout, "error (non-fatal): need to add a circle, "
                    "but no arc layer available\n");
    return;
  }

  w = ug.addNewWire(arc_layer, current_region);
  assert(w != NULL);

  v1 = ug.addNewVert(x, y, z, current_region);
  v2 = ug.addNewVert(x, y+r, z, current_region);
  v3 = ug.addNewVert(x, y+r, z, current_region);
  v4 = ug.addNewVert(x, y, z, current_region);
  
  assert(v1 != NULL && v2 != NULL && v3 != NULL && v4 != NULL);

  w->pushVert(v1);
  w->pushVert(v2);
  w->pushVert(v3);
  w->pushVert(v4);

  printf("CIRCLE HAS NAME %s\n", w->name);
  */

  ug_vert *v;
  StringPack sp;
  string type;

  if (arc_layer == NULL) {
    fprintf(stdout, "error (non-fatal): need to add a circle, "
                    "but no arc layer available\n");
    return;
  }

  type = string("circle");
  sp.addString(string("type"), type);
  sp.addString(string("layer"), layer->get_name());
  sp.addDouble(string("radius"), r);

  v = ug.addNewNamedVert((char*)sp.serialize().c_str(),
                         x, y, z, current_region);
}

/*
  (cx, cy, cz) is the center, (mx, my, mz) is an endpoint of the major axis,
  ratio is the ratio of minor axis to major axis,
  a1 and a2 are the start and stop angles respectively
*/
void UG_Builder::addAnonymousEllipse(double cx, double cy, double cz,
                                     double mx, double my, double mz,
                                     double ratio, double a1, double a2,
                                     ug_layer *layer)
{
  ug_vert *v;
  StringPack sp;
  string type;

  if (arc_layer == NULL) {
    fprintf(stdout, "error (non-fatal): need to add a circle, "
                    "but no arc layer available\n");
    return;
  }

  type = string("ellipse");
  sp.addString(string("type"), type);
  sp.addString(string("layer"), layer->get_name());
  /*
  sp.addDouble(string("cx"), cx);
  sp.addDouble(string("cy"), cy);
  sp.addDouble(string("cz"), cz);
  */
  sp.addDouble(string("mx"), mx);
  sp.addDouble(string("my"), my);
  sp.addDouble(string("mz"), mz);
  sp.addDouble(string("ratio"), ratio);
  sp.addDouble(string("a1"), a1);
  sp.addDouble(string("a2"), a2);

  v = ug.addNewNamedVert((char*)sp.serialize().c_str(),
                         cx, cy, cz, current_region);
}

void UG_Builder::addTextVertex(string text, string layer_name,
                               double ix, double iy, double iz)
{
  ug_vert *v;
  StringPack sp;

  sp.addString(string("text"), text);
  sp.addString(string("layer"), layer_name);

  // adding these is redundant
  sp.addDouble(string("x"), ix);
  sp.addDouble(string("y"), iy);
  sp.addDouble(string("z"), iz);

  v = ug.addNewNamedVert((char*)sp.serialize().c_str(), ix, iy, iz,
                         current_region);
  //v = ug.addNewNamedVert((char*)text.c_str(), ix, iy, iz, current_region);

  assert(v != NULL);
}


// this imposes a restriction that each line be 255 bytes at most
//   (including newline)
void UG_Builder::readLayerFile(char *filename)
{
  
  FILE *f;
  char *r;
  char linebuf[LAYER_LINE_LEN];
  int l, n, s;
  char *layername, *yesno, *colornum;
  unsigned int i;
  unsigned int cur_color, max_color = 0;
  vector<double> rv, gv, bv;
  ug_layer *layer;

  f = fopen(filename, "r");
  
  // should throw an exception..
  assert(f != NULL);

  while (!feof(f)) {
    r = fgets(linebuf, LAYER_LINE_LEN, f);
    if (r == NULL) break;
    n = strlen(linebuf);

    // change this to an exception..
    // this tells us that the line is not too long
    assert(linebuf[n-1] == '\n');
    
    s = 0;
    l = readAlphanumToken(linebuf, s, n); // s gets changed
    assert(l+s <= n-1);
    linebuf[l+s] = '\0';
    layername = linebuf+s;
    s = l+s+1;

    l = readAlphanumToken(linebuf, s, n); // s gets changed
    assert(l+s <= n-1);
    linebuf[l+s] = '\0';
    yesno = linebuf+s;
    s = l+s+1;

    l = readAlphanumToken(linebuf, s, n); // s gets changed
    assert(l+s <= n-1);
    linebuf[l+s] = '\0';
    colornum = linebuf+s;

    printf("got this info: %s %s %s\n", layername, yesno, colornum);

    // conditions on second token
    assert(yesno[1]=='\0');
    assert(yesno[0]=='Y' || yesno[0]=='N' || yesno[0]=='A');

    // colornum should be a non-negative integer
    for (i=0; colornum[i] != '\0'; i++)
      assert(NUMBER(colornum[i]));

    if (yesno[0] == 'Y') {
      cur_color = atoi(colornum);
      if (cur_color > max_color) max_color = cur_color;
      layer = new ug_layer(layername, 0.0, 0.0, 0.0, cur_color);
      assert(layer != NULL);
      valid_layers.add(layer);

    } else if (yesno[0] == 'A') {
      // this is the arc/ellipse/circle layer..
      // the DXF should not have anything on this layer, BUT we will
      //   use it for arcs/circles/ellipses

      cur_color = atoi(colornum);
      if (cur_color > max_color) max_color = cur_color;

      // we should only have one arc_layer
      assert(arc_layer==NULL);

      // it would be better to add the arc_layer only if we actually get
      //   arcs, and just create it here; but, this is OK for now
      arc_layer = ug.addNewLayer(layername, 0.0, 0.0, 0.0, cur_color);
      assert(arc_layer != NULL);

      // we do not add this to valid_layers, because the DXF should NOT
      //   be explicitly putting things into it
      // however, we put it in now just to get the layer/color working
      valid_layers.add(arc_layer);

    } 
    /* else if (yesno[0] == 'S') {
      // this is the space names layer

      cur_color = atoi(colornum);
      if (cur_color > max_color) max_color = cur_color;

      // we should only have one name_layer
      assert(name_layer==NULL);

      name_layer = ug.addNewLyer(layername, 0.0, 0.0, 0.0, cur_color);
      assert(name_layer != NULL);

      // again, eventually we don't want this added here yet
      valid_layers.add(name_layer);
      } */

  }

  fclose(f);

  // now, generate the colors we need
  for (i=0; i < max_color; i++) {
    rv.push_back(fabs(random() % 256) / 255.0);
    gv.push_back(fabs(random() % 256) / 255.0);
    bv.push_back(fabs(random() % 256) / 255.0);
  }

  // now fill in the colors based on the indices
  for (i=0; i < valid_layers.size; i++) {
    layer = valid_layers.at(i);
    layer->r = rv.at(layer->c - 1);
    layer->g = gv.at(layer->c - 1);
    layer->b = bv.at(layer->c - 1);
  }

  printf("END\n");
}

void UG_Builder::writeUGFile(const string &filename)
{
  ug_write(ug, (char*)filename.c_str());
}

ug_layer *UG_Builder::getLayer(string layer_name)
{
  return valid_layers.get_named((char*)layer_name.c_str());
}

// reads an alphanumeric token, returns the length of it
// start is where we should begin reading, and gets changed
//   to the pos. of the beginning of the token
// len is the full length of the buffer
int UG_Builder::readAlphanumToken(char *b, int &start, int len)
{
  char *bp;

  bp = b+start;

  // first read in whitespace
  while (bp-b < len) {
    if (!WHITESPACE(*bp)) break;
    bp++;
  }

  start = bp-b;
  
  // now read till the end of the token
  while (bp-b < len) {
    if (WHITESPACE(*bp)) break;
    bp++;
  }

  return bp-(b+start);
}

